//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "InjEct.h"

ClassImp(InjEct)

/////////////////////////////////////////////////////////////////////////
InjEct::InjEct(Streams* aStream, const string aPattern, const int aVerbose): InjRea(aPattern, aVerbose), GwollumPlot("INJ"){ 
/////////////////////////////////////////////////////////////////////////

  // for plotting
  SetGridx();  SetGridy();
  
  InStream = aStream;
  
  // check detector
  if(InStream->GetDetIndex()<0){
    cerr<<"InjEct::InjEct: the input stream must be a LV detector"<<endl;
    status_OK=false;
  }
  
  // check frequency
  if(InStream->GetNativeFrequency()<16){
    cerr<<"InjEct::InjEct: the input stream's native sampling frequency must be larger than 16Hz"<<endl;
    status_OK=false;
  }
    
  // dummy containers
  window=GetTukeyWindow(10,2);
  hplus  = new TGraph(1); hplus->SetName("hplus");
  hcross = new TGraph(1); hcross->SetName("hcross");
  hdet   = new TGraph(1); hdet->SetName("hdet");

  // real containers
  UpdateNativeSamplingFrequency();
}

/////////////////////////////////////////////////////////////////////////
InjEct::~InjEct(void){
/////////////////////////////////////////////////////////////////////////
  if(Verbose>1) cout<<"InjEct::~InjEct"<<endl;
  delete hplus;
  delete hcross;
  delete hdet;
  delete window;
}

/////////////////////////////////////////////////////////////////////////
void InjEct::UpdateNativeSamplingFrequency(void){
/////////////////////////////////////////////////////////////////////////
  if(InStream->GetNativeFrequency()==sampling) return;
  delete window;
  delete hplus;
  delete hcross;
  delete hdet;

  sampling=InStream->GetNativeFrequency();
  
  // injection size = Max( 1s, 10-sigma )
  int injsize=TMath::Max(1,(int)ceil(10.0*GetInjectionSigmaMax())*sampling);

  // tukey window
  window=GetTukeyWindow(injsize,(int)((double)injsize/10.0));
  
  // injection waveform (+)
  hplus  = new TGraph(injsize); hplus->SetName("hplus");
  hplus->GetXaxis()->SetNoExponent();
  hplus->SetTitle("hplus");
  hplus->SetMarkerColor(2);
  hplus->SetLineColor(2);
  
  // injection waveform (x)
  hcross = new TGraph(injsize); hcross->SetName("hcross");
  hcross->SetTitle("hcross");
  hcross->SetMarkerColor(4);
  hcross->SetLineColor(4);

  // injection waveform projected in the detector (antenna factors)
  hdet   = new TGraph(injsize); hdet->SetName("hdet");
  hdet->SetTitle("hdet");
  hdet->SetMarkerColor(5);
  hdet->SetLineColor(5);

  return;
}

/////////////////////////////////////////////////////////////////////////
bool InjEct::Inject(const int aDataSize, double *aData, const double aTimeStart){
/////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"InjEct::Inject: the InjEct object is corrupted"<<endl;
    return false;
  }
  if(aData==NULL){
    cerr<<"InjEct::Inject: Input vector is null"<<endl;
    return false;
  }
  if(InStream->GetNativeFrequency()!=sampling){
    cerr<<"InjEct::Inject: the native sampling frequency has changed --> cannot inject"<<endl;
    return false;
  }

  // block size
  double start = aTimeStart-hplus->GetN()/2/sampling;
  double stop  = aTimeStart+(aDataSize+hplus->GetN()/2)/sampling;

  // injection index range
  int injindex_start = GetFirstInjectionIndex(start,stop);
  int injindex_stop  = GetLastInjectionIndex(start,stop);

  // nothing to inject  
  if(injindex_start<0) return true;
  if(injindex_stop<0)  return true;

  double tlocal;// local gps time
  int dataindex;

  for(int i=injindex_start; i<=injindex_stop; i++){

    // genrate waveforms
    if(!GenerateWaveforms(i)) return false;

    // add injection waveform
    for(int t=0; t<hdet->GetN(); t++){
      tlocal=hdet->GetX()[t]+inj_time;
      if(tlocal<aTimeStart) continue;
      dataindex=(tlocal-aTimeStart)*sampling;
      if(dataindex<0) continue; //before data vector
      if(dataindex>=aDataSize) break; //after data vector
      aData[dataindex]+=hdet->GetY()[t];
    }
  }


  return true;  
}

/////////////////////////////////////////////////////////////////////////
bool InjEct::PlotInjection(const int aInjIndex){
/////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"InjEct::PlotInjection: the Inject object is corrupted"<<endl;
    return false;
  }
  if(aInjIndex<0||aInjIndex>=GetNInjections()){
    cerr<<"InjEct::PlotInjection: the injection index #"<<aInjIndex<<" does not exist"<<endl;
    return false;
  }

  // wavforms
  if(!GenerateWaveforms(aInjIndex)) return false;

  // plot
  Draw(hplus, "APL");
  Draw(hcross, "PLsame");
  Draw(hdet, "PLsame");
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool InjEct::GenerateWaveforms(const int aInjIndex){
/////////////////////////////////////////////////////////////////////////

  // load injection
  if(!LoadInjection(aInjIndex)) return false;

  // compute hplus and hcross
  if(!GetInjectionType().compare("SineGauss")) GenerateSineGauss();
  else return false;

  // project onto the detector --> hdet
  Project();

  return true;
}


/////////////////////////////////////////////////////////////////////////
void InjEct::GenerateSineGauss(void){
/////////////////////////////////////////////////////////////////////////
  // This function was taken from lalsimulation/src/LALSimBurst.c

  // "peak" amplitudes of plus and cross
  const double h0plus  = GetSineGaussh0plus();
  const double h0cross = GetSineGaussh0cross();

  // parameters
  double sigma  = 2*inj_sigma*inj_sigma;
  double factor = 2.0 * TMath::Pi() * inj_f0;
  
  // make waveforms centered on 0
  double t;
  for(int i=0; i<hplus->GetN(); i++){
    t = (double)(i-hplus->GetN()/2.0)/(double)sampling;
    hplus->SetPoint(i,t, h0plus * exp(-t*t/sigma) * cos(factor*t));
    hcross->SetPoint(i,t, h0cross * exp(-t*t/sigma) * sin(factor*t));
  }
  
  return;
}

/////////////////////////////////////////////////////////////////////////
void InjEct::Project(void){
/////////////////////////////////////////////////////////////////////////

  double tgeo; // geocentric time
  double tlocal; // local time
  double fplus, fcross;// antenna factors
 
  // fill hdet
  for(int i=0; i<hplus->GetN(); i++){

    // geocentric time
    tgeo = inj_time + (i-hplus->GetN()/2.0)/(double)sampling;
    
    // local time
    tlocal=InStream->GetLocalTime(inj_ra, inj_dec, tgeo);

    // antenna factors
    InStream->GetDetAMResponse(fplus, fcross, inj_ra, inj_dec, inj_psi, GreenwichMeanSiderealTime(tlocal));
    
    // populate hdet
    hdet->SetPoint(i,tlocal-inj_time, fplus * hplus->GetY()[i] + fcross * hcross->GetY()[i]);
    
  }

  return;
}

////////////////////////////////////////////////////////////////////////////////////
double* InjEct::GetTukeyWindow(const int aSize, const int aFractionSize){
////////////////////////////////////////////////////////////////////////////////////

  int FracSize_2=aFractionSize/2;
  double *Window = new double [aSize];
  
  double factor = TMath::Pi()/(double)FracSize_2;
  
  for (int i=0; i<FracSize_2; i++)
    Window[i] = 0.5*(1+TMath::Cos(factor*(double)(i-FracSize_2)));
  for (int i=FracSize_2; i<aSize-FracSize_2; i++)
    Window[i] = 1.0;
  for (int i=aSize-FracSize_2; i<aSize; i++)
    Window[i] = 0.5*(1+TMath::Cos(factor*(double)(i-aSize+FracSize_2)));
 
  return Window;
}
