//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "Spectrum.h"

ClassImp(Spectrum)

////////////////////////////////////////////////////////////////////////////////////
static int compare_double(const void *p1, const void *p2){
////////////////////////////////////////////////////////////////////////////////////
  double x1 = *(const double *)p1;
  double x2 = *(const double *)p2;
  return (x1 > x2) - (x1 < x2);
}

////////////////////////////////////////////////////////////////////////////////////
Spectrum::Spectrum(const int aSpectrumSize, const int aDataDuration, const int aDataSamplingFrequency, const int aVerbosity){ 
////////////////////////////////////////////////////////////////////////////////////

  // save parameters
  double DataDuration   = fabs(aDataDuration);
  DataSamplingFrequency = fabs(aDataSamplingFrequency);
  fVerbosity            = aVerbosity;
  status_OK             = true;
  int DataSize          = DataDuration*DataSamplingFrequency;

  //***************** CHECKS *****************
  if(!DataSamplingFrequency){
    cerr<<"Spectrum::Spectrum: the sampling frequency does not make sense"<<endl;
    status_OK=false;
  }
  if(DataSamplingFrequency%2){
    cerr<<"Spectrum::Spectrum: the sampling frequency must be an even number"<<endl;
    status_OK=false;
  }
  if(aSpectrumSize<2){
    cerr<<"Spectrum::Spectrum: the spectrum size does not make sense"<<endl;
    status_OK=false;
  }
  if(aSpectrumSize%2){
    cerr<<"Spectrum::Spectrum: the spectrum size must be an even number"<<endl;
    status_OK=false;
  }
  if(DataSize<2*aSpectrumSize){
    cerr<<"Spectrum::Spectrum: the input data size is too small to compute a PSD with this resolution"<<endl;
    status_OK=false;
  }
  //******************************************

  // FFT
  fft_psd = new fft(2*aSpectrumSize, "FFTW_MEASURE", "r2c");
 
  // nominal number of subsegments
  if(status_OK){
    nSubSegments[0] = DataSize/fft_psd->GetSize_t();
    nSubSegments[1] = (DataSize-fft_psd->GetSize_t()/2)/fft_psd->GetSize_t();
  }
  else{
    nSubSegments[0]=0;
    nSubSegments[1]=0;
  }
  
  // allocate memory
  for(int t=0; t<2; t++){
    flag_buffer[t] = new bool [nSubSegments[t]];
    PSD_buffer[t]  = new double* [nSubSegments[t]];
    for(int s=0; s<nSubSegments[t]; s++)
      PSD_buffer[t][s] = new double [fft_psd->GetSize_f()];
  }
  PSD        = new double [fft_psd->GetSize_f()];// one-sided
  PSD_var    = new double [fft_psd->GetSize_f()];// one-sided
  logPSD     = new double [fft_psd->GetSize_f()];// one-sided
  f_PSD      = new double [fft_psd->GetSize_f()];// one-sided
  HannWindow = GetHannWindow(fft_psd->GetSize_t());
   
  // frequency vector
  for(int i=0; i<fft_psd->GetSize_f(); i++)
    f_PSD[i]=i*(double)DataSamplingFrequency/(double)fft_psd->GetSize_t();
  
  // interpolator
  acc = gsl_interp_accel_alloc();
  interp_psd = gsl_spline_alloc(gsl_interp_linear, fft_psd->GetSize_f());
  
  // init all
  Reset();

  // print info
  if(fVerbosity>1){
    cout<<"************* Spectrum::Spectrum *************"<<endl;
    cout<<"Input data sampling frequency = "<<DataSamplingFrequency<<" Hz"<<endl;
    cout<<"PSD vector size               = "<<fft_psd->GetSize_f()<<" points"<<endl;
    cout<<"PSD resolution                = "<<(double)DataSamplingFrequency/(double)fft_psd->GetSize_t()<<" Hz"<<endl;
    cout<<"Number of even samples        = "<<nSubSegments[0]<<endl;
    cout<<"Number of odd samples         = "<<nSubSegments[1]<<endl;
    cout<<"**********************************************"<<endl;
  }
}

////////////////////////////////////////////////////////////////////////////////////
Spectrum::~Spectrum(void){
////////////////////////////////////////////////////////////////////////////////////
  if(fVerbosity>1) cout<<"Spectrum::~Spectrum"<<endl;
  delete HannWindow;
  delete f_PSD;
  delete PSD;
  delete PSD_var;
  delete logPSD;
  delete fft_psd;
  for(int t=0; t<2; t++){
    for(int s=0; s<nSubSegments[t]; s++) delete PSD_buffer[t][s];
    delete PSD_buffer[t];
    delete flag_buffer[t];
  }
  gsl_spline_free(interp_psd);
  gsl_interp_accel_free(acc);
}

////////////////////////////////////////////////////////////////////////////////////
void Spectrum::Reset(void){
////////////////////////////////////////////////////////////////////////////////////
  for(int t=0; t<2; t++){
    stop_buffer[t]=-1;
    for(int s=0; s<nSubSegments[t]; s++) flag_buffer[t][s]=false;
  }
  for(int i=0; i<fft_psd->GetSize_f(); i++){ PSD[i]=0.0; PSD_var[i]=0.0; logPSD[i]=0.0; }
}

////////////////////////////////////////////////////////////////////////////////////
double Spectrum::GetPower(const double aFrequency){
////////////////////////////////////////////////////////////////////////////////////
  double logPower = -1.0;
  if(!status_OK){
    cerr<<"Spectrum::GetPower: the Spectrum object is corrupted"<<endl;
    return logPower;
  }

  if(aFrequency<=0) return PSD[0]; // return DC
  
  // extrapolate below
  if(aFrequency<=f_PSD[1]){
    logPower = logPSD[1] + (logPSD[2]-logPSD[1])/(f_PSD[2]-f_PSD[1]) * (aFrequency - f_PSD[1]);// linearly
    //if(logPower<=0) return PSD[1];// constant
  }
  // extrapolate above
  else if(aFrequency>f_PSD[fft_psd->GetSize_f()-1]){
    logPower = (logPSD[fft_psd->GetSize_f()-1]-logPSD[fft_psd->GetSize_f()-2])/(f_PSD[fft_psd->GetSize_f()-1]-f_PSD[fft_psd->GetSize_f()-2]) * (aFrequency - f_PSD[fft_psd->GetSize_f()-1]) + logPSD[fft_psd->GetSize_f()-1];
    //if(logPower<=0) return PSD[fft_psd->GetSize_f()-1];// constant
  }
  // interpolate
  else{
    logPower=gsl_spline_eval(interp_psd, aFrequency, acc);// linearly
  }

  if(logPower==0.0) return 0.0; 
  return exp(logPower);
}

////////////////////////////////////////////////////////////////////////////////////
double Spectrum::GetAmplitude(const double aFrequency){
////////////////////////////////////////////////////////////////////////////////////
  double logPower = -1.0;
  if(!status_OK){
    cerr<<"Spectrum::GetAmplitude: the Spectrum object is corrupted"<<endl;
    return logPower;
  }

  if(aFrequency<=0) return sqrt(abs(PSD[0])); // return DC

  // extrapolate below
  if(aFrequency<=f_PSD[1]){
    logPower = logPSD[1] + (logPSD[2]-logPSD[1])/(f_PSD[2]-f_PSD[1]) * (aFrequency - f_PSD[1]);// linearly
    //if(logPower<=0) return sqrt(PSD[1]);// constant
  }
  // extrapolate above
  else if(aFrequency>f_PSD[fft_psd->GetSize_f()-1]){
    logPower = (logPSD[fft_psd->GetSize_f()-1]-logPSD[fft_psd->GetSize_f()-2])/(f_PSD[fft_psd->GetSize_f()-1]-f_PSD[fft_psd->GetSize_f()-2]) * (aFrequency - f_PSD[fft_psd->GetSize_f()-1]) + logPSD[fft_psd->GetSize_f()-1];
    //if(logPower<=0) return sqrt(PSD[fft_psd->GetSize_f()-1]);// constant
  }
  // interpolate
  else{
    logPower=gsl_spline_eval(interp_psd, aFrequency, acc);// linearly
  }
  
  if(logPower==0.0) return 0.0;
  return exp(logPower/2.0);
}

////////////////////////////////////////////////////////////////////////////////////
bool Spectrum::AddData(const int aDataSize, const double *aData, const int aDataStart){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"Spectrum::AddData: the Spectrum object is corrupted"<<endl;
    return false;
  }

  // number of new PSD subsegments
  int n_sub[2];
  n_sub[0] = aDataSize/fft_psd->GetSize_t();
  n_sub[1] = (aDataSize-fft_psd->GetSize_t()/2)/fft_psd->GetSize_t();
  if(n_sub[0]<1){
    cerr<<"Spectrum::AddData: the input data vector is too small to compute a PSD with this resolution"<<endl;
    return false;
  }
  
  // loop over even and odd sub-segments
  int index_start;
  for(int t=0; t<2; t++){// even and odd
    if(!nSubSegments[t]) continue; // not to divide by 0
    for(int s=0; s<n_sub[t]; s++){
      index_start=aDataStart + (2*s+t)*fft_psd->GetSize_t()/2;
      
      // fill subsegment data vector
      for(int i=0; i<fft_psd->GetSize_t(); i++) fft_psd->SetRe_t(i,aData[index_start+i]*HannWindow[i]);
    
      // fft-forward it
      fft_psd->Forward();

      // get power for this sub-segment
      // normalize power spectrum to give correct units:
      // 2/Tspec = 2f_s/N (periodogram)
      // /f_s / f_s (FFT squared)
      stop_buffer[t]=(stop_buffer[t]+1)%nSubSegments[t];
      for(int i=0; i<fft_psd->GetSize_f(); i++)
	PSD_buffer[t][stop_buffer[t]][i]=2.0*fft_psd->GetNorm2_f(i)/(double)fft_psd->GetSize_t()/(double)DataSamplingFrequency;
      flag_buffer[t][stop_buffer[t]]=true;
      
    }
  }
 
  if(fVerbosity) cout<<"Spectrum::LoadData: compute PSD"<<endl;
  ComputeMedianMean();
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Spectrum::LoadData(const int aDataSize, const double *aData, const int aDataStart){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"Spectrum::LoadData: the Spectrum object is corrupted"<<endl;
    return false;
  }
  
  Reset();
  return AddData(aDataSize,aData,aDataStart);
}

////////////////////////////////////////////////////////////////////////////////////
void Spectrum::ComputeMedianMean(void){
////////////////////////////////////////////////////////////////////////////////////

  // count number of active buffers
  int size_odd=0, size_evn=0;
  for(int s=0; s<nSubSegments[0]; s++) if(flag_buffer[0][s]) size_evn++;
  for(int s=0; s<nSubSegments[1]; s++) if(flag_buffer[1][s]) size_odd++;

  // warning: non-nominal PSD estimation
  if(fVerbosity>1){
    if(size_evn!=nSubSegments[0] || size_odd!=nSubSegments[1])
      cout<<"Spectrum::ComputeMedianMean: the PSD estimation is not nominal: "<<size_evn<<"/"<<size_odd<<" subPSDs are computed instead of "<<nSubSegments[0]<<"/"<<nSubSegments[1]<<endl;
  }
  
  // allocate memory
  double *sorted_odd = new double [size_odd];
  double *sorted_evn = new double [size_evn];
  double median_odd=0, median_evn=0;
  int sa;
  
  // loop over frequency bins
  for(int i=0; i<fft_psd->GetSize_f(); i++){

    // init variance
    PSD_var[i] = 0.0;

    // loop over even buffers
    sa=0;
    for(int s=0; s<nSubSegments[0]; s++){
      if(flag_buffer[0][s]){
	sorted_evn[sa]=PSD_buffer[0][s][i];
	PSD_var[i]+=PSD_buffer[0][s][i]*PSD_buffer[0][s][i];
	sa++;
      }
    }

    // loop over odd buffers
    sa=0;
    for(int s=0; s<nSubSegments[1]; s++){
      if(flag_buffer[1][s]){
	sorted_odd[sa]=PSD_buffer[1][s][i];
        PSD_var[i]+=PSD_buffer[0][s][i]*PSD_buffer[0][s][i];
	sa++;
      }
    }

    // sort PSD values
    qsort( sorted_odd, size_odd, sizeof(*sorted_odd), compare_double );
    qsort( sorted_evn, size_evn, sizeof(*sorted_evn), compare_double );

    // take median and correct median bias
    if(size_odd)
      median_odd=(sorted_odd[(size_odd-1)/2]+sorted_odd[size_odd/2])/2.0;
    else
      median_odd=0.0;
    if(size_evn)
      median_evn=(sorted_evn[(size_evn-1)/2]+sorted_evn[size_evn/2])/2.0;
    else
      median_evn=0.0;
    median_odd/=medianbiasfactor(size_odd);
    median_evn/=medianbiasfactor(size_evn);
    
    // average odd and even medians
    PSD[i]   = ((double)size_odd*median_odd+(double)size_evn*median_evn) / (double)(size_odd+size_evn);
    PSD_var[i] = PSD_var[i]/(double)(size_odd+size_evn)-PSD[i]*PSD[i];
    if(PSD[i]>0) logPSD[i]=log(PSD[i]); else logPSD[i]=0.0;
  }

  // smoothing
  //for(int i=2; i<fft_psd->GetSize_f()-1; i++)
  //PSD[i] = (PSD[i-1]+PSD[i]+PSD[i+1])/3.0;
  
  // clean memory
  delete sorted_odd;
  delete sorted_evn;
  
  // reset interpolator
  gsl_spline_init(interp_psd, f_PSD, logPSD, fft_psd->GetSize_f());
  gsl_interp_accel_reset(acc);

  return;
}

////////////////////////////////////////////////////////////////////////////////////
double* Spectrum::GetHannWindow(const int aSize){
////////////////////////////////////////////////////////////////////////////////////
  
  double *Window = new double [aSize];
  double factor = 2.0*TMath::Pi()/(double)(aSize-1);
  double norm = 0.0;
  
  for (int i=0; i<aSize; i++){
    Window[i] = 0.5 * (1.0-TMath::Cos(factor*(double)(i)));
    norm += Window[i]*Window[i];
  }

  // enforce unity RMS on the window
  for(int i=0; i<aSize; i++) Window[i]/=sqrt(norm/(double)aSize);

  return Window;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* Spectrum::GetPSD(void){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK) return NULL;
  TGraph *GPSD=new TGraph(fft_psd->GetSize_f(),f_PSD,PSD);
  GPSD->SetName("PSD");
  GPSD->SetTitle("Power spectrum density");
  return GPSD;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* Spectrum::GetPSD(const double aFrequencyMin, const double aFrequencyMax){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK) return NULL;
  TGraph *GPSD=new TGraph();
  GPSD->SetName("PSD");
  GPSD->SetTitle("Power spectrum density");
  int p=0;
  for(int i=1; i<fft_psd->GetSize_f(); i++){
    if(f_PSD[i]<aFrequencyMin) continue;
    if(f_PSD[i]>aFrequencyMax) break;
    GPSD->SetPoint(p,f_PSD[i],PSD[i]);
    p++;
  }
  return GPSD;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* Spectrum::GetASD(void){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK) return NULL;
  TGraph *GASD=new TGraph(fft_psd->GetSize_f()-1);
  GASD->SetName("ASD");
  GASD->SetTitle("Amplitude spectrum density");
  for(int i=1; i<fft_psd->GetSize_f(); i++) GASD->SetPoint(i-1,f_PSD[i],sqrt(PSD[i])); 
  return GASD;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* Spectrum::GetASD(const double aFrequencyMin, const double aFrequencyMax){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK) return NULL;
  TGraph *GASD=new TGraph();
  GASD->SetName("ASD");
  GASD->SetTitle("Amplitude spectrum density");
  int p=0;
  for(int i=1; i<fft_psd->GetSize_f(); i++){
    if(f_PSD[i]<aFrequencyMin) continue;
    if(f_PSD[i]>aFrequencyMax) break;
    GASD->SetPoint(p,f_PSD[i],sqrt(PSD[i]));
    p++;
  }
  return GASD;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* Spectrum::GetSubPSD(const int aParity, const int aIndex){
////////////////////////////////////////////////////////////////////////////////////
  if(aParity<0||aParity>1) return NULL;
  if(aIndex<0||aIndex>=nSubSegments[aParity]) return NULL;
  if(!flag_buffer[aParity][aIndex]) return NULL;// not produced
  
  TGraph *G = new TGraph(fft_psd->GetSize_f(),f_PSD,PSD_buffer[aParity][aIndex]);
  stringstream ss;
  if(aParity) ss<<"PSD_odd_"<<aIndex;
  else        ss<<"PSD_even_"<<aIndex;
  G->SetName(ss.str().c_str());
  G->SetTitle(ss.str().c_str());
  G->SetLineColor(12);
  return G;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* Spectrum::GetSubASD(const int aParity, const int aIndex){
////////////////////////////////////////////////////////////////////////////////////
  if(aParity<0||aParity>1) return NULL;
  if(aIndex<0||aIndex>=nSubSegments[aParity]) return NULL;
  if(!flag_buffer[aParity][aIndex]) return NULL;// not produced
  
  TGraph *G = new TGraph();
  stringstream ss;
  if(aParity) ss<<"ASD_odd_"<<aIndex;
  else        ss<<"ASD_even_"<<aIndex;
  G->SetName(ss.str().c_str());
  G->SetTitle(ss.str().c_str());
  G->SetLineColor(12);

  for(int i=1; i<fft_psd->GetSize_f(); i++) G->SetPoint(i-1,f_PSD[i],sqrt(PSD_buffer[aParity][aIndex][i]));
   
  return G;
}

////////////////////////////////////////////////////////////////////////////////////
bool Spectrum::WriteSubPSD(const string aOutFileName){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK) return false;

  TFile *fpsd = new TFile(aOutFileName.c_str(),"RECREATE");
  if(fpsd->IsZombie()) return false;
  fpsd->cd();
  TGraph *G;
  for(int t=0; t<2; t++){
    for(int s=0; s<nSubSegments[t]; s++){
      G = GetSubPSD(t, s);
      if(G==NULL) continue;
      G->Write();
      delete G;
    }
  }
  G = GetPSD();
  G->Write();
  delete G;
  fpsd->Close();
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Spectrum::WriteSubASD(const string aOutFileName){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK) return false;

  TFile *apsd = new TFile(aOutFileName.c_str(),"RECREATE");
  if(apsd->IsZombie()) return false;
  apsd->cd();
  TGraph *G;
  for(int t=0; t<2; t++){
    for(int s=0; s<nSubSegments[t]; s++){
      G = GetSubASD(t, s);
      if(G==NULL) continue;
      G->Write();
      delete G;
    }
  }
  G = GetASD();
  G->Write();
  delete G;
  apsd->Close();
  return true;
}

