//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __ReadAscii__
#define __ReadAscii__

#include "CUtils.h"

#define NCOLMAX 50

using namespace std;

/**
 * Parse text files with columns.
 * Many operations are then available to access and modify the data elements.
 * \author    Florent Robinet
 */
class ReadAscii{

 public:
  
  /**
   * @name Constructors and destructors
   @{
  */

  /**
   * Constructor of the ReadAscii class.
   * With this constructor, the text file is loaded and data are extracted. The text file must be composed of a fix number of data columns. This file should not contain more than 50 columns. Columns should be separated by white spaces. Empty lines and commented lines ('//', '#' or '\%') are ignored. If known, the format of each column should be provided in argument. The format syntax is the folowing: \"t;t;t\", where ';' separates each column definition and 't' refers to the type of data which is expected. Three data types are supported:
   * - 's' for a string type
   * - 'i' for a int type
   * - 'd' for a double/float type
   * \code{.cpp}
   * ReadAscii("/path/to/my/file.txt", "s;s;i;i");
   * \endcode
   *
   * This example means that the text file is composed of 4 columns, the 2 first ones are string fields and the 2 last ones are integer fields.
   *
   * If the input format is unknown, aformat should be set to "UNKNOWN" which it is by default. In that case, all columns are assumed to have a string format. This default format can be changed later on by using the SetFormat() function.
   *
   * Lines with a wrong number of columns (as defined by aformat or by the first line) are skipped.
   * @param afilename path to the text file
   * @param aformat string format to be expected in the file
   */
  ReadAscii(const string afilename, const string aformat="UNKNOWN");

  /**
   * Destructor of the ReadAscii class.
   */
  virtual ~ReadAscii(void);
 
  /**
     @}
  */

  /**
   * Returns the number of rows/lines.
   */
  inline int GetNLine(void){ return (int)mylines.size(); };

  /**
   * Returns the number of rows/lines.
   */
  inline int GetNRow(void){ return (int)mylines.size(); };

  /**
   * Returns the number of columns.
   */
  inline int GetNCol(void){ return (int)Vformat.size(); };

  /**
   * Dumps the content of a line as a string.
   * true is returned in case of success.
   * @param line returned line.
   * @param li selected line number (indexing starts at 0)
   */
  bool GetLine(string &line, const int li);

  /**
   * Dumps the content of an integer column in a vector.
   * true is returned in case of success.
   * @param col returned vector of integer data
   * @param co selected column number (indexing starts at 0)
   */
  bool GetCol(vector<int> &col, const int co);

  /**
   * Dumps the content of a double column in a vector.
   * true is returned in case of success.
   * @param col returned vector of double data
   * @param co selected column number (indexing starts at 0)
   */
  bool GetCol(vector<double> &col, const int co);

  /**
   * Dumps the content of a string column in a vector.
   * true is returned in case of success.
   * @param col returned vector of string data.
   * @param co selected column number (indexing starts at 0).
   */
  bool GetCol(vector<string> &col, const int co);

  /**
   * Returns the integer data value at a given line and column position.
   * true is returned in case of success.
   * @param element returned data value.
   * @param co selected column number (indexing starts at 0).
   * @param li selected line number (indexing starts at 0).
   */
  bool GetElement(int &element, const int li, const int co);

  /**
   * Returns the double data value at a given line and column position.
   * true is returned in case of success.
   * @param element returned data value.
   * @param co selected column number (indexing starts at 0).
   * @param li selected line number (indexing starts at 0).
   */
  bool GetElement(double &element, const int li, const int co);

  /**
   * Returns the string data value at a given line and column position.
   * true is returned in case of success.
   * @param element returned data value.
   * @param co selected column number (indexing starts at 0).
   * @param li selected line number (indexing starts at 0).
   */
  bool GetElement(string &element, const int li, const int co);

  /**
   * Defines a new column format ('s', 'd' or 'i').
   * The selected column is converted to a new format. Only one conversion is destructive: 'd' to 'i'. In that case, the doubles are recasted as integers.
   * true is returned in case of success.
   * @param co selected column number (indexing starts at 0).
   * @param aformat char new format ('s', 'd' or 'i').
   */
  bool SetFormat(const int co, const char aformat);

 private:

  // extract the elements of the files
  bool ExtractFormat(void); ///< extract format vector
  bool ExtractLines(void);  ///< extract lines
  bool ExtractCol(void);    ///< extract columns
  void GetMinMax(const int co); ///< extract min/max values

  string ffilename;           ///< input file name
  string fformat;             ///< input format string
  vector<string> Vformat;     ///< vector of formats
  vector <string> mylines;    ///< lines of the files
  vector <string> scol[NCOLMAX]; ///< column of string
  vector <double> dcol[NCOLMAX]; ///< column of double
  vector <int>    icol[NCOLMAX]; ///< column of int
  double colmin[NCOLMAX];        ///< minimum in a given column
  double colmax[NCOLMAX];        ///< maximum in a given column
};
 
#endif









